import {sleep} from 'k6';
import {SharedArray} from 'k6/data';
import http from 'k6/http';
import logging from 'k6/x/frostfs/logging';
import registry from 'k6/x/frostfs/registry';
import stats from 'k6/x/frostfs/stats';

import {newGenerator} from './libs/datagen.js';
import {parseEnv} from './libs/env-parser.js';
import {textSummary} from './libs/k6-summary-0.0.2.js';
import {uuidv4} from './libs/k6-utils-1.4.0.js';

parseEnv();

const obj_list = new SharedArray('obj_list', function() {
  return JSON.parse(open(__ENV.PREGEN_JSON)).objects;
});

const container_list = new SharedArray('container_list', function() {
  return JSON.parse(open(__ENV.PREGEN_JSON)).containers;
});

const read_size = JSON.parse(open(__ENV.PREGEN_JSON)).obj_size;
const summary_json = __ENV.SUMMARY_JSON || '/tmp/summary.json';

// Select random HTTP endpoint for current VU
const http_endpoints = __ENV.HTTP_ENDPOINTS.split(',');
const http_endpoint =
    http_endpoints[Math.floor(Math.random() * http_endpoints.length)];
const log = logging.new().withField('endpoint', http_endpoint);

const registry_enabled = !!__ENV.REGISTRY_FILE;
const obj_registry =
    registry_enabled ? registry.open(__ENV.REGISTRY_FILE) : undefined;

const duration = __ENV.DURATION;

if (!!__ENV.METRIC_TAGS) {
  stats.setTags(__ENV.METRIC_TAGS)
}

const scenarios = {};

const write_vu_count = parseInt(__ENV.WRITERS || '0');
const generator = newGenerator(write_vu_count > 0);
if (write_vu_count > 0) {
  scenarios.write = {
    executor: 'constant-vus',
    vus: write_vu_count,
    duration: `${duration}s`,
    exec: 'obj_write',
    gracefulStop: '5s',
  }
}

const read_vu_count = parseInt(__ENV.READERS || '0');
if (read_vu_count > 0) {
  scenarios.read = {
    executor: 'constant-vus',
    vus: read_vu_count,
    duration: `${duration}s`,
    exec: 'obj_read',
    gracefulStop: '5s',
  }
}

export const options = {
  scenarios,
  setupTimeout: '5s',
};

export function setup() {
  const total_vu_count = write_vu_count + read_vu_count;

  console.log(`Pregenerated containers:       ${container_list.length}`);
  console.log(`Pregenerated read object size: ${read_size}`);
  console.log(`Pregenerated total objects:    ${obj_list.length}`);
  console.log(`Reading VUs:                   ${read_vu_count}`);
  console.log(`Writing VUs:                   ${write_vu_count}`);
  console.log(`Total VUs:                     ${total_vu_count}`);

  const start_timestamp = Date.now()
  console.log(
      `Load started at:               ${Date(start_timestamp).toString()}`)
}

export function teardown(data) {
  if (obj_registry) {
    obj_registry.close();
  }
  const end_timestamp = Date.now()
  console.log(
      `Load finished at:              ${Date(end_timestamp).toString()}`)
}

export function handleSummary(data) {
  return {
    'stdout': textSummary(data, {indent: ' ', enableColors: false}),
    [summary_json]: JSON.stringify(data),
  };
}

export function obj_write() {
  if (__ENV.SLEEP_WRITE) {
    sleep(__ENV.SLEEP_WRITE);
  }

  const container =
      container_list[Math.floor(Math.random() * container_list.length)];

  const payload = generator.genPayload();
  const data = {
    field: uuidv4(),
    // Because we use `file` wrapping and it is not straightforward to use
    // streams here,
    // `-e STREAMING=1` has no effect for this scenario.
    file: http.file(payload.bytes(), 'random.data'),
  };

  const resp = http.post(`http://${http_endpoint}/upload/${container}`, data);
  if (resp.status != 200) {
    log.withFields({status: resp.status, cid: container}).error(resp.error);
    return;
  }
  const object_id = JSON.parse(resp.body).object_id;
  if (obj_registry) {
    obj_registry.addObject(container, object_id, '', '', payload.hash());
  }
}

export function obj_read() {
  if (__ENV.SLEEP_READ) {
    sleep(__ENV.SLEEP_READ);
  }

  const obj = obj_list[Math.floor(Math.random() * obj_list.length)];
  const resp =
      http.get(`http://${http_endpoint}/get/${obj.container}/${obj.object}`);
  if (resp.status != 200) {
    log.withFields({status: resp.status, cid: obj.container, oid: obj.object})
        .error(resp.error);
  }
}
